#
#
# Name:   difflts.py - quantifying the distance between two non-equivalent LTSs.
# Author: Gwen Salaun
#
################################################################################

 # useful for real division

##
# A label is a string / message / action
class Label:

    def __init__(self,message):
        self._message=message.lower() # why lower ?

    def getLabel(self):
        return self._message

    def printLabel(self):
        print(self._message)

##
# A state is an identifier
class State:

    def __init__(self,name):
        self._name=name

    def getName(self):
        return self._name

    def printState(self):
        print(self._name, end=' ')

##
# A transition is a tuple (source state, label, target state)
class Transition:
    def __init__(self,src,label,tgt):
        self._src=src
        self._tgt=tgt
        self._label=label

    def getSource(self):
        return self._src

    def getTarget(self):
        return self._tgt

    def getLabel(self):
        return self._label

    def printTransition(self):
        print(self._src.getName()+"-"+self._label.getLabel()+"->"+self._tgt.getName(), end=' ')

##
# An LTS is a tuple (name, init, S, T)
class LTS:
    def __init__(self,name,init):
        self._name = name
        self._s0 = init
        self._S = []
        self._T = []

    def addState(self, state):
        if not state.getName() in [s.getName() for s in self._S]:
            self._S.append(state)

    def addTransition(self,transition):
        self._T.append(transition)

    def getName(self):
        return self._name

    def getStates(self):
        return self._S

    def getTransitions(self):
        return self._T

    def getInit(self):
        return self._s0

    # Builds alphabet
    def buildAlphabet(self):
        alpha=[]
        for t in self._T:
            if (t.getLabel().getLabel()!="i"):
                res=t.getLabel().getLabel()+"_"+t.getLabel().getDirection()
                if not(res in alpha):
                    alpha.append(res)
        return alpha

    # Returns the list of transitions with sid as source
    #def getTransitionsIdent(self,sid):
    #    subT=[]
    #    for t in self._T:
    #        if (sid==t.getSource().getName()):
    #           subT.append(t)
    #    return subT

    def printLTS(self):
        print("states {", end=' ')
        for s in self._S:
            s.printState(),
        print("}")
        print("transitions {", end=' ')
        for t in self._T:
            t.printTransition(),
        print("}")

    # this function returns all transitions with state s as source
    def getOutgoingTransitions(self,s):
        trans=[]
        for t in self._T:
            if (t.getSource().getName()==s.getName()):
                trans.append(t)
        return trans

    # this function take a (target state) and a partition, and returns
    # the block of that state
    def computeBlock(self,s,partition):
        for blockc in partition: # each block is a couple (block id, state list)
            for state in blockc[1]:
                if (s.getName()==state.getName()):
                    return blockc[0]

    # this function returns the states corresponding to a block name
    def getBlockStates(self,block,partition):
        for blockc in partition: # each block is a couple (block id, state list)
            if (block==blockc[0]):
                return blockc[1]

    # this function checks in a signature (label, block name) belongs to a
    #  list of signatures
    def isInSignatures(self,s,signatures):
        res=False
        for sig in signatures:
            if (s[0].getLabel()==sig[0].getLabel()) and (s[1]==sig[1]):
                res=True
        return res

    # this function removes a signature from a list and returns the modified list
    def removeSignature(self,s,signatures):
        cpt=0
        ind=-1
        for sig in signatures:
            if (s[0].getLabel()==sig[0].getLabel()) and (s[1]==sig[1]):
                ind=cpt
            cpt=cpt+1
        # removes the entry at index 'ind' in the list
        if (ind>=0):
            del signatures[ind]
        return signatures

    # this function compares two signatures where a signature is
    #  a list of couples (label, block name)
    def sameSignature(self,s1,s2):
        import copy
        s2bis = copy.deepcopy(s2)
        #print "BEGIN sameSignature"
        res=True
        for s in s1: # s is a couple (label, block name)
            # check if s is in s2
            #s[0].printLabel()
            #print s[1]
            if not(self.isInSignatures(s,s2bis)):
                res=False
            else:
                #print "FOUND"
                # remove s from s2
                self.removeSignature(s,s2bis)
        #print "END sameSignature"
        return res and (s2bis==[])

    # useful for the partition algorithm
    # partition is a set of blocks
    # where each block is a couple (block ident, set of states)
    def signature(self,state,partition):
        res=[]
        trans=self.getOutgoingTransitions(state)
        # print "signature"
        # state.printState()
        for t in trans:
            # t.printTransition()
            res.append((t.getLabel(),self.computeBlock(t.getTarget(),partition)))
        return res


    # this function computes the state signatures for a set of states
    # given a specific partition
    def computeSignatures(self,states,partition):
        res=[]
        for s in states:
            res.append((s,self.signature(s,partition)))
        return res

    # this function gets a list of state signatures and prints them
    # each signature has a a state and a list of couple (label, target state)
    def printSignatures(self,signatures):
        print("state signatures")
        for sres in signatures:
            print("state ", sres[0].getName(), " -> ", end=' ')
            for c in sres[1]:
                print((c[0].getLabel(), c[1]), end=' ')
            print("")

    # this function takes a block, a partition, and an LTS
    # and splits the block into several ones if they have different signatures
    # indname (integer) is useful to give fresh block names
    def split (self,block,partition,indname):
        #print "SPLIT function"
        # partition=[("B0",lts.getStates())]
        states=self.getBlockStates(block,partition)
        signatures=self.computeSignatures(states,partition)
        #self.printSignatures(signatures)
        # res is a list of signatures (signature=(state, list of (label, target state)))

        # we split the states into different blocks
        # the result (resblocks) is a list of tuples, each tuple consists of
        # a (new) block name, the set of states and the corresponding signature

        resblocks=[]
        for sres in signatures:
            state=sres[0] # we want to add that state in resblocks at the right place
            #state.printState()
            if self.existSignature(sres[1],resblocks):
                #print "EXISTING"
                resblocks=self.addStateExistingBlock(state,sres[1],resblocks)
            else: # we create a new block
                #print "NEW BLOCK"
                resblocks.append(("B"+str(indname),[state],sres[1]))
                indname=indname+1
            #print resblocks
        return (indname,resblocks)

    # this function checks if a signature is an existing block where
    #  - sig is a signature
    #  - resblocks is a list of tuple (block name, the set of states, corresponding signature)
    def existSignature(self,sig,resblocks):
        found=False
        for block in resblocks:
            if self.sameSignature(sig,block[2]):
                found=True
        return found

    # this function adds a state to an existing block where
    #  - state is the state to be added
    #  - sig is its signature
    #  - resblocks is a list of tuple (block name, the set of states, corresponding signature)
    def addStateExistingBlock(self,state,sig,resblocks):
        for block in resblocks:
            if self.sameSignature(sig,block[2]):
                block[1].append(state) # CAUTION: traversing and modifying at the same time ??
        return resblocks

    # checks if two lists of states are identical
    def sameStates(self,ls1,ls2):
        import copy
        ls2bis=copy.deepcopy(ls2)
        res=True
        for state in ls1:
            if not(self.isInState(state,ls2bis)):
                res=False
            else:
                self.removeState(state,ls2bis)
        return res and (ls2bis==[])

    # checks if a state is in a list of states
    def isInState(self,state,ls):
        for s in ls:
            if (state.getName()==s.getName()):
                return True
        return False

    # this function removes a state from a list of states
    def removeState(self,state,ls):
        cpt=0
        ind=-1
        for s in ls:
            if (state.getName()==s.getName()):
                ind=cpt
                break
            cpt=cpt+1
        if (ind>=0):
            del ls[ind]
        return ls

    # this function checks whether a block nam belongs to a partition
    def isInPartition(self,states,partition):
        for bc in partition: # bc is a couple (block name, list of states)
            if (self.sameStates(states,bc[1])):
                return True
        return False

    # this function compares two partitions
    # a partition is a list of couples (block name, list of states)
    def samePartition(self,partition,partition2):
        import copy
        partition2bis = copy.deepcopy(partition2)
        res=True
        #print "PARTITION"
        #print partition
        #print partition2
        for bc in partition: # bc is a couple (block name, list of states)
            # check if bc is in partition2 (just checking the block name)
            if not(self.isInPartition(bc[1],partition2bis)):
                res=False
            else:
                # remove bc from partition2
                self.removeBlock(bc[1],partition2bis)
        #print "PARTITION RESULT", res and (partition2bis==[])
        return res and (partition2bis==[])


    # this function removes a block from a partition given a block name
    def removeBlock(self,states,partition):
        cpt=0
        ind=-1
        for cblock in partition:
            if (self.sameStates(states,cblock[1])):
                ind=cpt
            cpt=cpt+1
        if (ind>=0):
            del partition[ind]
        return partition

    # this function computes the union of a partition and of the split result
    def unionPartition(self,partition,partition2):
        res=[]
        for bc in partition:
            if not(self.isInPartition(bc[1],partition2)):
                res.append(bc)
        return res+partition2

    # this functions converts the result of the split function
    #  -- list of tuples (block name, list of states, signature) --
    #  to the partition format, ie, list of couples (block name, list of states)
    def convertToPartition(self,resblocks):
        res=[]
        for entry in resblocks:
            res.append((entry[0],entry[1]))
        return res

    # this function prints out a partition (list of couples (block name, state list))
    def printPartition(self,partition):
        #print "partition"
        for bc in partition:
            print("Block", bc[0], "{", end=' ')
            for s in bc[1]:
                print(s.getName(), end=' ')
            print("}")


    # this function builds the coarsest partition (with fewest blocks)
    #  for an LTS
    def buildPartition(self):
        import copy
        # initial partition
        partition=[("B0",self.getStates())]
        indname=1
        res=self.computeSignatures(self.getStates(),partition)
        prevPartition=[]
        #print "before while"
        cpt=0
        while not(self.samePartition(partition,prevPartition)):
            #print "while (beginning)"
            cpt=cpt+1
            #self.printPartition(partition)
            prevPartition = copy.deepcopy(partition)
            for cblock in prevPartition:
                # we call split function
                splitres=self.split(cblock[0],partition,indname)
                #for entry in splitres[1]:
                #    print entry[0], " ",
                #    for s in entry[1]:
                #        s.printState(),
                #    print ""
                indname=splitres[0]
                resblocks=splitres[1] # list of tuples (block name, list of states, signature)
                partition3=self.convertToPartition(resblocks) # convert to partition format
                partition2=self.removeBlock(cblock[1],partition)# remove block[1] from partition
                partition=self.unionPartition(partition2,partition3) # merge P and result of split

        return partition

##
# The 3 following classes are useful for Measure2 in the Comparator class

##
# An extended state is an (identifier, a boolean bisim, a distance dist,
# a distance to the init state).
# The boolean 'bisim' indicates if that state is bisimilar to another one
# The distance indicates the minimum distance to a bisim state if not(bisim)
#  zero otherwise
class XState:

    def __init__(self,name,bisim,distbis,distinit):
        self._name=name
        self._bisim=bisim
        self._distbis=distbis
        self._distinit=distinit

    def getName(self):
        return self._name

    def getBisim(self):
        return self._bisim

    def isBisim(self):
        return self._bisim

    def getDistanceBis(self):
        return self._distbis

    def setDistanceBis(self,distbis):
        self._distbis=distbis

    def getDistanceInit(self):
        return self._distinit

    def setDistanceInit(self,distinit):
        self._distinit=distinit

##
# An extended transition is a tuple (source state, label, target state)
class XTransition:
    def __init__(self,src,label,tgt):
        self._src=src
        self._tgt=tgt
        self._label=label

    def getSource(self):
        return self._src

    def getTarget(self):
        return self._tgt

    def getLabel(self):
        return self._label

    def equals(self,t):
        return (self._src.getName()==t.getSource().getName()) and \
        (self._tgt.getName()==t.getTarget().getName()) and \
        (self._label.getLabel()==t.getLabel().getLabel())


##
# An extended LTS is a tuple (name, init, XS, XT)
# TODO: use inheritance
class XLTS:
    def __init__(self,name,init,states,transitions):
        self._name = name
        self._s0 = init
        self._XS = states
        self._XT = transitions

    def addState(self, state):
        self._XS.append(state)

    def addTransition(self,transition):
        self._XT.append(transition)

    def getStates(self):
        return self._XS

    def getTransitions(self):
        return self._XT

    def getState(self,statename):
        for state in self._XS:
            if state.getName()==statename:
                return state

    def getInitState(self):
        return self._s0

    # this function returns all transitions with state s as source
    def getOutgoingXTransitions(self,s):
        trans=[]
        for t in self._XT:
            if (t.getSource().getName()==s.getName()):
                trans.append(t)
        return trans

    # this function returns all transitions with state s as target
    def getIncomingXTransitions(self,s):
        trans=[]
        for t in self._XT:
            if (t.getTarget().getName()==s.getName()):
                trans.append(t)
        return trans

    # this function checks whether a state is initial or not
    def isInit(self, xs):
        res=False
        if xs.getName()==self._s0.getName():
            res=True
        return res

##
# This class is used for loading LTSs from aut files and compare them
#  using partition refinement algorithms
class Comparator:
    def __init__(self,f1,f2):
            self._filename1=f1
            self._filename2=f2

    def loadLTS(self,filename):
        import re
        from subprocess import call

        nfile=filename+".aut"
        nfileaux=filename+"_aux.aut"

        string=open(nfile).read()
        string=string[4:]
        new_str=re.sub('[^a-zA-Z0-9\n\.\ \_]', '', string)
        open(nfileaux,'w').write(new_str)

        # reads all the file contents and storing into data
        lines=tuple(open(nfileaux, "r"))
        data=[[0 for x in range(3)] for x in range(len(lines))]
        for j in range(0, len(lines)):
            data[j]=[s.strip() for s in lines[j].split(' ')]
		# print data

        # preprocessing on states
        setStates=list()
        setStates.append(data[0][0])
        for i in range(1,len(data)):
            if data[i][0] not in setStates:
                setStates.append(data[i][0])
            if data[i][2] not in setStates:
                setStates.append(data[i][2])

        # builds the LTS object
        newLTS=LTS(filename,State("s"+str(setStates[0])))
        for s in setStates:
            newLTS.addState(State("s"+str(s)))
        for i in range(1, len(data)):
            newLTS.addTransition(Transition(State("s"+str(data[i][0])),Label(data[i][1]),State("s"+str(data[i][2]))))

        call("rm "+nfileaux, shell=True)
        return newLTS

    # this function builds the union of 2 LTSs
    # and return a 3rd one
    def unionLTS(self,lts_1,lts_2):
        import copy
        lts1=copy.deepcopy(lts_1)
        lts2=copy.deepcopy(lts_2)
        init = lts1.getInit()
        reslts=LTS(lts1.getName()+lts2.getName(),init)
        # we rename state identifiers for ensuring uniqueness
        for s in lts1.getStates():
            reslts.addState(State("lts1_"+s.getName()))
        for s in lts2.getStates():
            reslts.addState(State("lts2_"+s.getName()))
        for t in lts1.getTransitions():
            t2=Transition(State("lts1_"+t.getSource().getName()),t.getLabel(),State("lts1_"+t.getTarget().getName()))
            reslts.addTransition(t2)
        for t in lts2.getTransitions():
            t2=Transition(State("lts2_"+t.getSource().getName()),t.getLabel(),State("lts2_"+t.getTarget().getName()))
            reslts.addTransition(t2)
        return reslts

    # this function removes a transition from a list of transitions
    def removeTransition(self,transition,lt):
        cpt=0
        ind=-1
        for t in lt:
            if (t.equals(transition)):
                ind=cpt
                break
            cpt=cpt+1
        if (ind>=0):
            del lt[ind]
        return lt

    def checkBisim(self,lts,init1,init2,partition):
        sid1="lts1_"+init1.getName()
        sid2="lts2_"+init2.getName()
        for bc in partition:
            ct=0            # if this counter reaches 2, it means that
            for s in bc[1]: # both init states are in the same block
                if (s.getName()==sid1):
                    ct=ct+1
                if (s.getName()==sid2):
                    ct=ct+1
            if (ct>=2):
                return True
            else:
                ct=0
        return False

    # Measure 1: number of bisimilar states in both LTSs and in the global LTS
    #  (numbers and percentages)
    # Takes as input the global LTS and the resulting partition
    def measure1(self,lts,partition):
        notBisLTS1=0 # number of non-bisimilar states in LTS1
        notBisLTS2=0 # number of non-bisimilar states in LTS2
        # a block with states belonging to only one LTS contains non-bis states
        for bc in partition:
            ctLTS1=0
            ctLTS2=0
            for s in bc[1]:
                if (s.getName()[:4]=="lts1"):
                    ctLTS1=ctLTS1+1
                if (s.getName()[:4]=="lts2"):
                    ctLTS2=ctLTS2+1
            if (ctLTS1>0) ^ (ctLTS2>0):
                notBisLTS1=notBisLTS1+ctLTS1
                notBisLTS2=notBisLTS2+ctLTS2
        return (notBisLTS1,notBisLTS2)
        #print "LTS1:", notBisLTS1, "not bis out of ",

    def checkBis(self,state,pref,partition):
        sid=pref+state.getName()
        for bc in partition:
            foundInBlock=False
            ctLTS1=0
            ctLTS2=0
            for s in bc[1]:
                if (sid==s.getName()):
                    foundInBlock=True
                if (s.getName()[:4]=="lts1"):
                    ctLTS1=ctLTS1+1
                if (s.getName()[:4]=="lts2"):
                    ctLTS2=ctLTS2+1
            if foundInBlock:
                return ((ctLTS1>0) and (ctLTS2>0)) # true if at least one state from both LTSs
        return False # not found :(

    # This function takes a state and a set of xstates, and return the
    #  corresponding xstate
    def getXState(self,state,xstates):
        for xs in xstates:
            if (state.getName()==xs.getName()):
                return xs

    # This function generates an XLTS from an LTS and set correctly all
    #  bisim attributes present in extended states
    def generateXLTS(self,lts,nb,partition):
        if (nb==1):
            pref="lts1_"
        else:
            pref="lts2_"
        # build the new list of states
        states=[]
        for s in lts.getStates():
            bis=self.checkBis(s,pref,partition)
            states.append(XState(s.getName(),bis,0,0))
        # build the new list of transitions
        transitions=[]
        for t in lts.getTransitions():
            source=self.getXState(t.getSource(),states)
            target=self.getXState(t.getTarget(),states)
            transitions.append(XTransition(source,t.getLabel(),target))
        # instantiate XLTS
        return XLTS(lts.getName(),self.getXState(lts.getInit(),states),states,transitions)

    # This function returns the successor states of an xstate
    def getSucc(self,xs,xlts):
        trans=xlts.getTransitions()
        succ=[]
        for t in trans:
            if (t.getSource().getName()==xs.getName()):
                succ.append((t.getLabel().getLabel(),t.getTarget()))
        return succ

    # This function searches in a list of couples (bool, dist) whether
    #  one of the booleans is set to True
    def isFoundState(self,cbooldist):
        for c in cbooldist:
            if c[0]:
                return True
        return False

    # This function searches in a list of couples (bool, dist) the shortest
    #  distance for booleans set to True
    def getMinDistance(self,cbooldist):
        res=100000000  # beurk
        for c in cbooldist:
            if c[0]: # bisim state found
                if (c[1]<res):
                    res=c[1]
        return res

    # This function measures the distance to the closest bisim state given an xstate
    #  and an XLTS
    def distanceToClosestBisState(self,xs,xlts):
        # We rely on a depth-first search algorithm with visited storage
        #  and min computation when moving back up
        res=self.distanceToClosestBisStateAUX(xs,xlts,[])
        return res[1]

    # This function measures the distance to the closest bisim state given an xstate
    #  and an XLTS, using a list of visited states and a current distance (?)
    def distanceToClosestBisStateAUX(self,xs,xlts,visited):
        if (xs.getName() in visited):
            #print "VISITED"
            return (False,0)
        if xs.getBisim():  # if this xstate is bisim
            #print "BISIM TRUE"
            return (True,0)
        else:
            #print "SEARCH IN DEPTH"
            visited.append(xs.getName())
            succ=self.getSucc(xs,xlts)
            #for s in succ:
            #    print s.getName()
            res=[] # a list of couples (bool, dist)
            for xs2bis in succ:
                xs2=xs2bis[1]
                #print "IN FOR"
                res.append(self.distanceToClosestBisStateAUX(xs2,xlts,visited))
            if self.isFoundState(res):
                #print "BISIM FOUND"
                return (True,self.getMinDistance(res)+1)
            else:
                #print "BISIM NOT FOUND"
                return (False,0)


    # Measure 2: distance (min / max / average) of non-bisim states to the
    #  closest bisim state for LTS1 / LTS2 / global LTS
    def measure2(self,lts1,lts2,partition):
        # we first convert both LTSs into XLTSs with correct bisim parameters
        xlts1=self.generateXLTS(lts1,1,partition)
        xlts2=self.generateXLTS(lts2,2,partition)

        xstates1=xlts1.getStates()
        for xs in xstates1:
            d=self.distanceToClosestBisState(xs,xlts1)
            xs.setDistanceBis(d)

        #print "XLTS1"
        #for xs in xlts1.getStates():
        #    print xs.getName(),
        #    print xs.getBisim(),
        #    print xs.getDistance()

        xstates2=xlts2.getStates()
        for xs in xstates2:
            d=self.distanceToClosestBisState(xs,xlts2)
            xs.setDistanceBis(d)

        # average / min / max for first LTS
        dmin1=100000000 # beurk
        dmax1=0
        dsum1=0
        dct1=0 # number of non-bisim states
        for xs in xlts1.getStates():
            if not(xs.getBisim()):
                dct1=dct1+1
                d=xs.getDistanceBis()
                dsum1=dsum1+d
                if (d<dmin1):
                    dmin1=d
                if (d>dmax1):
                    dmax1=d
        if (dct1!=0):
            davg1=dsum1/float(dct1)
        else:
            dmin1=0
            davg1=0

        # TODO : use a method to avoid duplication
        # average / min / max for second LTS
        dmin2=100000000 # beurk
        dmax2=0
        dsum2=0
        dct2=0 # number of non-bisim states
        for xs in xlts2.getStates():
            if not(xs.getBisim()):
                dct2=dct2+1
                d=xs.getDistanceBis()
                dsum2=dsum2+d
                if (d<dmin2):
                    dmin2=d
                if (d>dmax2):
                    dmax2=d
        if (dct2!=0):
            davg2=dsum2/float(dct2)
        else:
            dmin2=0
            davg2=0

        return ((dmin1,dmax1,davg1),(dmin2,dmax2,davg2))

    # Checks whether a couple of state is already in a set of couples of states
    def isInToExplore(self,s1,s2,toexplore):
        for c in toexplore:
            if ((s1.getName()==c[0].getName()) and (s2.getName()==c[1].getName())):
                return True
        return False

    # Measure 3: computes first the product between two LTSs and
    #  then quantifies this common part wrt the size of both LTSs
    def measure3(self,lts1,lts2):
        # we use a list of couple of states to be explored
        # this list is initialised with the initial states
        toexplore=[(lts1.getInit(),lts2.getInit())]
        alreadyhandled=[] # list of already added state identifiers
        s0name=lts1.getInit().getName()+lts2.getInit().getName()
        newLTS=LTS(lts1.getName()+lts2.getName(),State(s0name))

        while (toexplore!=[]):
            cs=toexplore[0] # head of the list
            toexplore=toexplore[1:] # tail of the list
            s1=cs[0] # state in LTS1
            s2=cs[1] # state in LTS2
            sourcename=s1.getName()+s2.getName()
            #print "ICI"
            #print sourcename
            alreadyhandled.append(sourcename)
            newLTS.addState(State(sourcename)) # add this new state
            transitions1=lts1.getOutgoingTransitions(s1) # all transitions outgoing from s1
            transitions2=lts2.getOutgoingTransitions(s2) # all transitions outgoing from s2
            for t1 in transitions1:
                for t2 in transitions2:
                    if (t1.getLabel().getLabel()==t2.getLabel().getLabel()):
                        #print "YES"
                        #print alreadyhandled
                        targetname=t1.getTarget().getName()+t2.getTarget().getName()
                        newLTS.addTransition(Transition(State(sourcename),Label(t1.getLabel().getLabel()),State(targetname)))
                        if not(targetname in alreadyhandled):
                            ts1=t1.getTarget()
                            ts2=t2.getTarget()
                            if not(self.isInToExplore(ts1,ts2,toexplore)):
                                toexplore.append((ts1,ts2))

                    # TODO : handle properly internal transitions (and exit transitions) !!

        return newLTS

    # This function checks whether two successive versions of a matrix
    #  exhibit closed enough values. In that case, we say the matrix stabilizes
    # delta is the error margin -> if the overall difference is less than delta
    # we say that the matrix is stable
    def checkStability(self,oldmatrix,matrix,delta):
        # measure the difference for all couples of states and divide
        # by the total number of couples to obtain the average difference
        nb1=len(matrix)
        keys=list(matrix.keys())
        nb2=len(matrix[keys[0]])
        vsum=0
        # print nb1, nb2, matrix
        for statename1 in list(matrix.keys()):
            for statename2 in list(matrix[statename1].keys()):
                # print statename1, statename2
                val=matrix[statename1][statename2]
                valold=oldmatrix[statename1][statename2]
                # print val,valold
                if (val>valold):
                    vsum=vsum+(val-valold)
                else:
                    vsum=vsum+(valold-val)
        averagediff=vsum/(nb1*nb2)
        # print nb1,nb2,averagediff
        if (averagediff<delta):
            return True
        else:
            return False

    # This function measures the distance from the initial state to a given xstate
    def distanceFromInitState(self,xs,xlts):
        # We rely on a depth-first search algorithm with visited storage
        #  and min computation when moving back up
        res=self.distanceFromInitStateAUX(xlts.getInitState(),xs,xlts,[])
        return res[1]

    # This function measures the distance from the init state to a given xstate
    #  in an XLTS, using a list of visited states
    def distanceFromInitStateAUX(self,xcur,xs,xlts,visited):
        if (xcur.getName() in visited):
            #print "VISITED"
            return (False,0)
        if xcur.getName()==xs.getName():  # we have reached the searched state
            #print "BISIM TRUE"
            return (True,0)
        else:
            #print "SEARCH IN DEPTH"
            visited.append(xcur.getName())
            succ=self.getSucc(xcur,xlts)
            #for s in succ:
            #    print s.getName()
            res=[] # a list of couples (bool, dist)
            for xs2bis in succ:
                #print "IN FOR"
                xs2=xs2bis[1]
                res.append(self.distanceFromInitStateAUX(xs2,xs,xlts,visited))
            if self.isFoundState(res):
                #print "BISIM FOUND"
                return (True,self.getMinDistance(res)+1)
            else:
                #print "BISIM NOT FOUND"
                return (False,0)

    # Counts the number of matching transitions (both directions)
    # This version works for deterministic LTSs only
    # (no 2 transitions going out from a state with same label)
    def matchingTransitions(self,trans1,trans2):
        ct1=0
        for t1 in trans1:
            l1=t1.getLabel().getLabel()
            for t2 in trans2:
                l2=t2.getLabel().getLabel()
                #print l1,l2
                if (l1==l2):
                    ct1=ct1+1
                    break
        nbt1=len(trans1)
        if (nbt1==0):
            res1=1 # if no transitions, this is perfect match
                   # (ok for one direction, but if in both, this is undefined)
        else:
            res1=(ct1/nbt1)
        ct2=0
        for t2 in trans2:
            l2=t2.getLabel().getLabel()
            for t1 in trans1:
                l1=t1.getLabel().getLabel()
                if (l1==l2):
                    ct2=ct2+1
                    break
        nbt2=len(trans2)
        if (nbt2==0):
            res2=1
        else:
            res2=(ct2/nbt2)

        if ((nbt1==0) and (nbt2==0)):
            #print "MOINS UN"
            return -1 # no transitions to compare (undefined)
        else:
            return (res1+res2)/2

    # Counts the number of matching transitions (both directions)
    # This version works for non-deterministic LTSs
    # (2 transitions going out from a state with same label is supported)
    # In that case, we can match the label only once
    def matchingTransitionsND(self,trans1,trans2):
        import copy
        # first direction
        ct1=0
        trans1_copy=copy.deepcopy(trans1)
        trans2_copy=copy.deepcopy(trans2)
        for t1 in trans1_copy:
            l1=t1.getLabel().getLabel()
            for t2 in trans2_copy:
                l2=t2.getLabel().getLabel()
                #print l1,l2
                if (l1==l2):
                    ct1=ct1+1
                    # remove t2 from trans2_copy
                    self.removeTransition(t2,trans2_copy)
                    break
        nbt1=len(trans1)
        if (nbt1==0):
            res1=1 # if no transitions, this is perfect match
                   # (ok for one direction, but if in both, this is undefined)
        else:
            res1=(ct1/nbt1)

        # second direction
        ct2=0
        trans1_copy=copy.deepcopy(trans1)
        trans2_copy=copy.deepcopy(trans2)
        for t2 in trans2_copy:
            l2=t2.getLabel().getLabel()
            for t1 in trans1_copy:
                l1=t1.getLabel().getLabel()
                if (l1==l2):
                    ct2=ct2+1
                    # remove t1 from trans1_copy
                    self.removeTransition(t1,trans1_copy)
                    break
        nbt2=len(trans2)
        if (nbt2==0):
            res2=1
        else:
            res2=(ct2/nbt2)

        if ((nbt1==0) and (nbt2==0)):
            #print "MOINS UN"
            return -1 # no transitions to compare (undefined)
        else:
            return (res1+res2)/2

    # compares the nature of states
    def sameStateTypes(self,xs1,xlts1,xs2,xlts2):
        res=0
        if xlts1.isInit(xs1) and xlts2.isInit(xs2):
            res=1
        if not(xlts1.isInit(xs1)) and not(xlts2.isInit(xs2)):
            res=1
        return res

    # compares the similarity of neighbourhoods
    def compareNeighbourhoods(self,xs1,xlts1,xs2,xlts2,matrix):
        # we start with outgoing transitions
        resout=0
        nbcombiout=0
        out1=xlts1.getOutgoingXTransitions(xs1)
        out2=xlts2.getOutgoingXTransitions(xs2)
        for t1 in out1:
            for t2 in out2:
                #print t1.getTarget().getName()
                s1=t1.getTarget()
                s2=t2.getTarget()
                # we compare only bisim states because non-bisim states
                # do not have any value in the matrix
                if not(s1.isBisim()) and not(s2.isBisim()):
                    resout=resout+matrix[s1.getName()][s2.getName()]
                    nbcombiout=nbcombiout+1

        # we then compute similarity for incoming transitions
        resinc=0
        nbcombiinc=0
        inc1=xlts1.getIncomingXTransitions(xs1)
        inc2=xlts2.getIncomingXTransitions(xs2)
        for t1 in inc1:
            for t2 in inc2:
                s1=t1.getSource()
                s2=t2.getSource()
                if not(s1.isBisim()) and not(s2.isBisim()):
                    resinc=resinc+matrix[s1.getName()][s2.getName()]
                    nbcombiinc=nbcombiinc+1

        if ((nbcombiout+nbcombiinc)>0):
            res=((resout+resinc)/(nbcombiout+nbcombiinc))
            #print res
            return res
        else:
            return 1 # it means that we cannot discriminate on this criterion

    # prints the matrix
    def printMatrix(self,matrix):
        CRED = '\033[91m'
        CGREY = '\033[90m'
        CVIOLET = '\33[35m'
        CEND = '\033[0m'
        mat=list(matrix.values())[0]
        print("\t", end=' ')
        for statename in list(mat.keys()):
            print(statename, end=' ')
            print("\t", end=' ')
        print("")
        for statename1 in list(matrix.keys()):
            print(statename1, end=' ')
            print("\t", end=' ')
            for statename2 in list(matrix[statename1].keys()):
                val=round(matrix[statename1][statename2],2)
                if (val>=0.6) and (val<0.8):
                    print(CVIOLET, end='')
                elif (val>=0.8):
                    print(CRED, end='')
                else:
                    print(CGREY, end='')
                print(val, end=' ')
                print(CEND, end='')
                print("\t", end=' ')
            print("")


    # This function searches the shortest path in a list of paths
    def getShortestTrace(self,cboollist):
        res=100000000  # beurk
        tr=[]
        for c in cboollist:
            if c[0]: # bisim state found
                if (len(c[1])<res):
                    res=len(c[1])
                    tr=c[1]
        return tr


    # This function computes the set of shortest traces to the closest bisim state
    #  given an xstate and an XLTS
    def computeTracesToClosestBisState(self,xs,xlts):
        # We rely on a depth-first search algorithm with visited storage
        #  and min computation when moving back up
        res=self.computeTracesToClosestBisStateAUX(xs,xlts,[],[])
        # print(res)
        return res[1]

    #
    def computeTracesToClosestBisStateAUX(self,xs,xlts,visited,solution):
        if (xs.getName() in visited):
            # print("VISITED")
            return (False,solution)
        if xs.getBisim():  # if this xstate is bisim
            # print("BISIM TRUE")
            # print(solution)
            return (True,solution)
        else:
            # print("SEARCH IN DEPTH")
            visited.append(xs.getName())
            succ=self.getSucc(xs,xlts)
            #for s in succ:
            #    print s.getName()
            res=[] # a list of couples (bool, path/list of labels)
            for xs2bis in succ:
                xs2=xs2bis[1]
                #print "IN FOR"
                import copy
                sol2=copy.deepcopy(solution)
                # sol2=solution
                sol2.append(xs2bis[0])
                #print(solution)
                #print(sol2)
                res.append(self.computeTracesToClosestBisStateAUX(xs2,xlts,visited,sol2))
                #print("RES ")
                #print(res)

            if self.isFoundState(res):
                # print("BISIM FOUND")
                tr=self.getShortestTrace(res)
                # print(tr)
                return (True,tr) # TODO: extend to ALL shortest traces
            else:
                # print("BISIM NOT FOUND")
                return (False,solution)


    # returns two values, the number of common elements times 2 and the number
    # of all elements
    # Remark: in its current version, we increment common even if case of repetion of an action
    def comparePaths(self,sp1,sp2):
        common=0
        lc1=[]
        for act1 in sp1:
            if act1 in sp2:
                common=common+1
                lc1.append(act1)
        lc2=[]
        for act2 in sp2:
            if act2 in sp1:
                lc2.append(act2)
        # check order
        ordered=0
        i=0
        for a1 in lc1:
            if (a1==lc2[0]):
                ordered=ordered+1
                i=i+1

        allelem=len(sp1)+len(sp2)
        if ((common*2)>allelem):
            return (allelem,allelem,ordered)
        else:
            return (common*2,allelem,ordered)




    # Measure 4: similarity of non-bisim states in both LTSs
    # Pre-condition: call this function only if both LTSs are not bisimilar
    def measure4(self,lts1,lts2,partition):

        # we first convert both LTSs into XLTSs with correct bisim parameters
        xlts1=self.generateXLTS(lts1,1,partition)
        xlts2=self.generateXLTS(lts2,2,partition)

        # we extend xlts with min distance to closest bisim state
        xstates1=xlts1.getStates()
        for xs in xstates1:
            d=self.distanceToClosestBisState(xs,xlts1)

            # TODO: compute the set of shortest traces to closest state
            # sp=self.computeTracesToClosestBisState(xs,xlts1)
            # print("plus cours chemin de "+xs.getName()+" a bisim state ")
            # print(sp)

            xs.setDistanceBis(d)
            d=self.distanceFromInitState(xs,xlts1)
            xs.setDistanceInit(d)
        xstates2=xlts2.getStates()
        for xs in xstates2:
            d=self.distanceToClosestBisState(xs,xlts2)
            xs.setDistanceBis(d)
            d=self.distanceFromInitState(xs,xlts2)
            xs.setDistanceInit(d)

        # we compute both sets of non-bisim states
        nbs1=[]
        for s in xlts1.getStates():
            if not(s.getBisim()):
                nbs1.append(s)
        nbs2=[]
        for s in xlts2.getStates():
            if not(s.getBisim()):
                nbs2.append(s)
        # we initialize the resulting matrix
        matrix={}
        for s1 in nbs1:
            matrix[s1.getName()]={}
            for s2 in nbs2:
                matrix[s1.getName()][s2.getName()]=1

        if (len(nbs1)==0) or (len(nbs2)==0):
            print("Warning: computation impossible, just one LTS with bisim states")
            return {}

        isStable=False
        # we iteratively refine the matrix values
        while not(isStable):
            import copy
            oldmatrix=copy.deepcopy(matrix)

            # we compare distances to the closest bisim state
            for statename1 in list(matrix.keys()):
                for statename2 in list(matrix[statename1].keys()):
                    # number of valide measures (at least 2 for nature and
                    # simil. of neighbouhood states)
                    valid=2

                    xs1=xlts1.getState(statename1)
                    xs2=xlts2.getState(statename2)
                    d1=xs1.getDistanceBis()
                    d2=xs2.getDistanceBis()

                    # comparison of distances to closest bisim states
                    if ((d1==0) and (d2==0)):
                        mdist1=0 # used when t  are no bisim states
                                 # in that case, this measure is not useful
                    else:
                        #print "M1"
                        valid=valid+1
                        mdist1=1-( abs(d1-d2) / max(d1,d2) )

                    d1=xs1.getDistanceInit()
                    d2=xs2.getDistanceInit()

                    # comparison of distance to init states
                    if ((d1==0) and (d2==0)): # both states are initial
                        valid=valid+1
                        mdist2=1 # this is a perfect match for that measure
                    else:
                        #print d1,d2,"M2"
                        valid=valid+1
                        mdist2=1-( abs(d1-d2) / max(d1,d2) )

                    # computation of path similarity TODO
                    sp1=self.computeTracesToClosestBisState(xs1,xlts1)
                    #print("plus court chemin de "+xs1.getName()+" a bisim state ")
                    #print(sp1)
                    sp2=self.computeTracesToClosestBisState(xs2,xlts2)
                    #print("plus court chemin de "+xs2.getName()+" a bisim state ")
                    #print(sp2)
                    if ((sp1!=[]) and (sp2!=[])):
                        valid=valid+1
                        ressp=self.comparePaths(sp1,sp2)
                        #print(ressp[0])
                        #print(ressp[1])
                        #print(ressp[2]) # result for order of actions
                        # we add a bonus when order is respected
                        if (ressp[2]==0):
                            mdist3=ressp[0]/ressp[1]
                        else:
                            bonus=ressp[2]/ressp[0]
                            w1=0.9
                            w2=0.1
                            mdist3=(ressp[0]/ressp[1])*w1 + bonus*w2
                        #print(mdist3)
                    else:
                        mdist3=0 # uncomparable if one set is empty

                    # comparison of outgoing transitions
                    out1=xlts1.getOutgoingXTransitions(xs1)
                    out2=xlts2.getOutgoingXTransitions(xs2)
                    # count matching outgoing transitions
                    mdist4=self.matchingTransitionsND(out1,out2)
                    if (mdist4>=0):
                        #print "M3"
                        #print mdist3
                        valid=valid+1
                    else:
                        mdist4=0

                    # comparison of incoming transitions
                    inc1=xlts1.getIncomingXTransitions(xs1)
                    inc2=xlts2.getIncomingXTransitions(xs2)
                    # count matching outgoing transitions
                    mdist5=self.matchingTransitionsND(inc1,inc2)
                    if (mdist5>=0):
                        #print "M4"
                        #print mdist4
                        valid=valid+1
                    else:
                        mdist5=0

                    # nature of states (init or not)
                    if self.sameStateTypes(xs1,xlts1,xs2,xlts2):
                        mdist6=1 # does not impact the similarity value
                    else:
                        mdist6=0 # reduces the similarity value

                    # similarity of neighbourhoods
                    mdist7=self.compareNeighbourhoods(xs1,xlts1,xs2,xlts2,oldmatrix)

                    # HERE USEFUL TO SEE THE DETAIL OF THE MATRIX
                    #print statename1, statename2
                    #print mdist1,mdist2,mdist3,mdist4,mdist5,mdist6

                    # compute the similarity of states xs1 and xs2
                    # we use uniform weights for values different of zero
                    w=1/valid
                    # print "WEIGHT",valid,w
                    matrix[statename1][statename2]= \
                    ((w*mdist1) + (w*mdist2) + (w*mdist3) + (w*mdist4) + (w*mdist5) + (w*mdist6) + (w*mdist7))

            # we update the condition
            isStable=self.checkStability(oldmatrix,matrix,0.05)
            self.printMatrix(matrix)
            #isStable=True # we do not iterate on the matrix YET

        return matrix

    # global measure on non-bisimilar states
    # average of best scores in both directions
    def computeGlobalSim(self,matrix):
        # computes a global measure from the matrix by computing average of max
        # values for each couples of states
        somme=0
        nb=0
        # it is worth noting we compute the global measure in both directions
        # because the matrix does not necessarily have the same number of entries
        for statename1 in list(matrix.keys()):
            maxi=0
            for statename2 in list(matrix[statename1].keys()):
                if (matrix[statename1][statename2]>maxi):
                    maxi=matrix[statename1][statename2]
                    #print statename1, statename2
            #print maxi
            somme=somme+maxi
            nb=nb+1
        mat=list(matrix.values())[0]
        for statename2 in list(mat.keys()):
            maxi=0
            for statename1 in list(matrix.keys()):
                if (matrix[statename1][statename2]>maxi):
                    maxi=matrix[statename1][statename2]
                    #print statename1, statename2
            #print maxi
            somme=somme+maxi
            nb=nb+1
        return round(somme/nb,2)


    # global measure on non-bisimilar states
    # average of all scores, one direction is enough because brut force
    def computeGlobalSimBrute(self,matrix):
        somme=0
        nb=0
        for statename1 in list(matrix.keys()):
            for statename2 in list(matrix[statename1].keys()):
                somme=somme+matrix[statename1][statename2]
                nb=nb+1
        return round(somme/nb,2)


    # global measure on non-bisimilar states
    # average of all scores higher than a threshold
    def computeGlobalSimThreshold(self,matrix, threshold):
        somme=0
        nb=0
        for statename1 in list(matrix.keys()):
            for statename2 in list(matrix[statename1].keys()):
                if (matrix[statename1][statename2]>threshold):
                    somme=somme+matrix[statename1][statename2]
                    nb=nb+1
        if (nb==0):
            return 0
        else:
            return round(somme/nb,2)


    # global measure on non-bisimilar states
    # average of only a percentage of the best scores
    # e.g., average of the 25% of best sores
    def computeGlobalSimPercentage(self,matrix,percentage):
        # transforms matrix to flat table
        tab=[]
        nb=0
        for statename1 in list(matrix.keys()):
            for statename2 in list(matrix[statename1].keys()):
                tab.append(matrix[statename1][statename2])
                nb=nb+1

        nbval=round((nb*percentage)/100)
        # print(nb)
        # print(percentage)
        # print(nbval)

        cpt=0
        somme=0
        tab=sorted(tab,reverse=True)
        while (cpt<nbval):
            # print(tab[cpt])
            somme=somme+tab[cpt]
            cpt=cpt+1

        if (cpt!=0):
            return round(somme/cpt,2)
        else:
            return 0


    # Compares two LTSs wrt. different quantitative measures
    def compareLTSs(self):
        lts1 = self.loadLTS(self._filename1)
        print("PRINT LTS1")
        lts1.printLTS()
        print("")
        # partition=lts1.buildPartition()
        # print "PRINT PARTITION"
        # lts1.printPartition(partition)

        lts2 = self.loadLTS(self._filename2)
        print("PRINT LTS2")
        lts2.printLTS()
        print("")

        lts=self.unionLTS(lts1,lts2)
        partition=lts.buildPartition()
        print("PRINT PARTITION")
        lts.printPartition(partition)
        print("")

        print("ARE THEY STRONGLY BISIMILAR ?", end=' ')
        bisimres=self.checkBisim(lts,lts1.getInit(),lts2.getInit(),partition)
        print(bisimres)
        print("")

        print("MEASURE 1: number of non-bisimilar states")
        res1=self.measure1(lts,partition)
        l1=len(lts1.getStates())
        l2=len(lts2.getStates())
        print("LTS1:", res1[0], "state(s) not bisim. out of", l1, "->",(res1[0]/float(l1))*100,"%")
        print("LTS2:", res1[1], "state(s) not bisim. out of", l2, "->", (res1[1]/float(l2))*100,"%")
        nonbisim=((res1[0]+res1[1])/float(l1+l2))*100
        print("Union LTS:", res1[0]+res1[1], "state(s) not bisim. out of", l1+l2, "->", nonbisim,"%")
        print("")

        res2=self.measure2(lts1,lts2,partition)
        print("MEASURE 2: distance of non-bisimilar states to bisimilar states")
        print("LTS1:", res2[0][0], "(min)", res2[0][1], "(max)", res2[0][2], "(avg)")
        print("LTS2:", res2[1][0], "(min)", res2[1][1], "(max)", res2[1][2], "(avg)")
        print("")

        res3=self.measure3(lts1,lts2)
        # res3.printLTS()
        nbstates=len(res3.getStates())
        print("MEASURE 3: quantification of the common part (synchronous product)")
        print("LTS1:", nbstates, "state(s) out of", l1, "in the product ->",(nbstates/float(l1))*100,"%")
        print("LTS2:", nbstates, "state(s) out of", l2, "in the product ->",(nbstates/float(l2))*100,"%")
        print("")

        print("MEASURE 4: similarity of non-bisimilar states in both LTSs")
        if not(bisimres):
            matrix=self.measure4(lts1,lts2,partition)
            # we print the matrix
            if (matrix!={}):
                self.printMatrix(matrix)
                # compute a global measure from the matrix by computing average of max
                # values for each couples of states (non-bisim states only)
                nonbisim=round(nonbisim,2)
                print("Global measure (brute average):",nonbisim,"% of non-bisimilar states, similar to",self.computeGlobalSimBrute(matrix)*100,"%")
                print("Global measure (average of best scores):",nonbisim,"% of non-bisimilar states, similar to",self.computeGlobalSim(matrix)*100,"%")
                resthres=self.computeGlobalSimThreshold(matrix,0.8)
                if (resthres!=0):
                    print("Global measure (average of best scores above threshold):",nonbisim,"% of non-bisimilar states, similar to",resthres*100,"%")
                resperc=self.computeGlobalSimPercentage(matrix,30)
                if (resperc!=0):
                    print("Global measure (average of best scores (percentage)):",nonbisim,"% of non-bisimilar states, similar to",resperc*100,"%")

        else:
            print("  -> there is no non-bisimilar states!")


if __name__ == '__main__':

    import sys
    from subprocess import call

    if len(sys.argv)!=3:
        print("Erroneous number of parameters !")
        print("Expected syntax: python difflts.py lts1 lts2 (aut extension assumed for filenames)")
    else:

        filename1=sys.argv[1]
        filename2=sys.argv[2]

        c = Comparator(filename1,filename2)
        c.compareLTSs()
